#!/usr/bin/python
"""
Foal - Familiar Open Application Launcher

Copyright (C) 2001 James ''Wez''Weatherall

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

Todo's

0.  Make icon-resizing work, either using 'convert' or GdkImlib
1.  Fix always-on-top code (somehow check pygtk version?)
2.  Some more standard icons...
3.  Integration with Dave Scott's Pointless-derived wm?
4.  Audible button clicks
5.  Integration with the iPAQ buttons
6.  Display of fscrib symbols used to launch same command
"""

try: import pylau
except ImportError: pass

import gtk
try: import GdkImlib
except ImportError: GdkImlib = None

import os
import sys
import string

ICON_SIZE = 16
def BUTTON_SIZE(): return ICON_SIZE+6
_icon_dirs = []

class RawWindow(gtk.GtkWindow):
    # - Base class for Launcher windows

    def __init__(self, size=None):
        gtk.GtkWindow.__init__(self, gtk.WINDOW_POPUP)
        if size:
            apply(self.set_usize, size)
        self.set_name("Foal")

    set_pos = gtk.GtkWindow.set_uposition

    def width(self): return self.get_window().width
    def height(self): return self.get_window().height

    def place_relative(self, widget, flags, size=None):
        # Place a new RawWindow relative to an existing widget
        # - Flags: 4/0 = horizontal/vertical
        pos = widget.get_allocation()
        wd = widget.get_window().toplevel
        pos = (pos[0]+wd.x, pos[1]+wd.y,
               pos[0]+pos[2]+wd.x, pos[1]+pos[3]+wd.y)
        if size: w,h = size
        else: w, h = self.width(), self.height()
        sw, sh = gtk.screen_width(), gtk.screen_height()
        if flags & 4:
            x=pos[0]-w
            y=(pos[1]+pos[3]-h)/2
            if x<0: x=pos[2]
        else:
            y=pos[1]-h
            x=(pos[0]+pos[2]-w)/2
            if y<0: y=pos[3]
        self.set_pos(max(0, min(x, sw-w)), max(0, min(y, sh-h)))

    def load_icon(self, icon, name):
        # - Load the icon for a named application
        if icon:
            for path in _icon_dirs:
                file = os.path.join(path, icon)
                try:
                    pixmap = gtk.GtkPixmap(self, file)
                except:
                    pixmap = None
                    if GdkImlib:
                        try:
                            image = GdkImlib.Image(file)
                            if image.render(ICON_SIZE, ICON_SIZE):
                                pixmap = i.make_pixmap()
                        except: pass
                if pixmap: return pixmap
        return self.text_icon(name)

    def text_icon(self, name):
        # - Create suitable text for an icon
        n=string.split(name)
        if not n: return gtk.GtkLabel("?")
        if len(n) > 1:
            txt=string.join(map(lambda x:x[0], n),"")
        else:
            if n[0][0] in string.uppercase:
                txt=n[0][0]
            else:
                txt=n[0]
        w=0 # - Magic number.  Is it "correct"?
        for i in range(len(txt)):
            w = w + self.get_style().font.width(txt[i])
            if w > ICON_SIZE:
                return gtk.GtkLabel(txt[:i])
        return gtk.GtkLabel(txt)

def gdk_intern_atom(name):
    import _gtk
    return _gtk.gdk_atom_intern(name, 1)

def set_gdk_win_prop(window, property, type,
                     format, mode, data):
    import _gtk, GDK
    atom = _gtk.gdk_atom_intern(property, 1)
    if not int(atom): return 0
    int_atom = _gtk.gdk_atom_intern(type, 1)
    if not int(int_atom): return 0
    if mode=="REPLACE":gdk_mode = GDK.PROP_MODE_REPLACE
    elif mode=="APPEND":gdk_mode = GDK.PROP_MODE_APPEND
    elif mode=="PREPEND":gdk_mode = GDK.PROP_MODE_PREPEND
    try:
        window.property_change(atom, int_atom, format, gdk_mode, data)
        return 1
    except: return 0

class LaunchButton(RawWindow):

    def __init__(self):

        # - Load the users preferences
        self.LoadPrefs()
        if self.__placement & 4:
            gtk.GtkWindow.__init__(self)
            self.set_usize(BUTTON_SIZE(), BUTTON_SIZE())
            self.realize()
            set_gdk_win_prop(self.get_window(), "_NET_WM_WINDOW_TYPE",
                             "ATOM", 32, "REPLACE",
                             [gdk_intern_atom("_NET_WM_WINDOW_TYPE_DOCK")])
        else:
            RawWindow.__init__(self, (BUTTON_SIZE(), BUTTON_SIZE()))

        # - Create the button
        button = gtk.GtkButton()
        button.set_name("FoalLauncher")
        self.__arrow = gtk.GtkArrow()
        self.__arrow.set(gtk.ARROW_UP, gtk.SHADOW_OUT)
        button.add(self.__arrow)
        button.connect("clicked", self.OnToggleMenu)
        self.add(button)

        # - Show the window
        self.connect("destroy", self.HideMenu)
        self.show_all()

        # - Place it
        self.place()

        self.__menu = None

    def place(self):
        if self.__placement & 4: return
            
        # - Select the location
        if self.__placement & 2:
            y = 0
        else:
            y = gtk.screen_height()-self.height()
        if self.__placement & 1:
            x = 0
        else:
            x = gtk.screen_width()-self.width()
        self.set_pos(x, y)
        
        # - Set the _WIN_LAYER property if possible
        if not set_gdk_win_prop(self.get_window(),
                                "_WIN_LAYER",
                                "INTEGER",
                                32,
                                "REPLACE",
                                [10L]):
            # - *** Can't set _WIN_LAYER - use crappy timeout hooks.
            print "foal:falling back to _raise hooks"
            self.__timeout = gtk.timeout_add(2000, self.OnTimeoutRaise)
            self.connect("destroy", self.OnDestroyRaise)

    def OnTimeoutRaise(self):
        self.get_window()._raise()
        return 1

    def OnDestroyRaise(self, widget):
        gtk.timeout_remove(self.__timeout)

    def LoadPrefs(self):
        # - Read the preferences

        global _icon_dirs

        # - Defaults
        self.__placement = 0 # - Bottom-Right
        _icon_dirs = [".", "icons"] # - Local directory & "icons" subdir
        self.__menuitems = []

        # - Load the RC files
        gtk.rc_reparse_all()

        # - Prefs files to load
        files = [os.path.join(os.environ["HOME"],
                              ".foal/foalrc"),
                 "/etc/foal/foalrc"]
        for name in files:
            try: self.ReadPrefs(open(name, "r"), [self.__menuitems])
            except:
                pass

    def GetItem(self, title):
        # - Find a menu item by title
        menu = self.__menuitems
        result = None
        for name in title:
            if not menu: return None
            for item in menu:
                if name == item.name:
                    result = item
                    if isinstance(item, LaunchMenuSubMenu): menu = item.items
                    else: menu = None
        return result

    def ReadPrefs(self, file, menu):
        # - Read a prefs file.

        # This includes icon directory(s),
        # icon program name substitutions and list of programs
        # and menus!
        # Also includes where to place the launcher button.

        global ICON_SIZE
        global _icon_dirs

        try:
            while 1:
                line = file.readline()
                if not line:
                    break
                line = string.strip(line)
                if not line or line[0] == "#":
                    continue

                try:
                    i = string.index(line, "(")
                    func = string.strip(line[:i])
                    data = eval(line[i:])
                    if type(data) != type(()):
                        data = (data,)
                except:
                    continue

                if func == "version":
                    pass
                elif func == "placement":
                    self.__placement = data[0]
                elif func == "icondir":
                    _icon_dirs.extend(list(data))
                elif func == "iconsize":
                    ICON_SIZE = data[0]
                elif func == "itemicon":
                    item = self.GetItem(data[0])
                    if item: item.icon = data[1]
                elif func == "startmenu":
                    if data[0]:
                        m = LaunchMenuSubMenu(data[0], data[1], [])
                        menu[0].append(m)
                        menu.insert(0, m.items)
                    else:
                        menu.insert(0, menu[0])
                elif func == "endmenu":
                    del menu[0]
                elif func == "menuitem":
                    menu[0].append(LaunchMenuItem(data[0], data[1], data[2]))
                elif func == "verifyitem":
                    submenu = [LaunchMenuItem(data[0]+":Yes", "yes_tick.xpm", data[2]),
                               LaunchMenuItem(data[0]+":NO", "no_cross.xpm", "")]
                    menu[0].append(LaunchMenuSubMenu(data[0],data[1],submenu))
                elif func == "include":
                    try: self.ReadPrefs(open(data[0], "r"), menu)
                    except: pass
        except:
            sys.stderr.write("warning:error reading Launcher prefs\n")

    def HideMenu(self, w=None):
        if self.__menu:
            self.__menu.destroy()
            self.__menu = None            
            self.__arrow.set(gtk.ARROW_UP, gtk.SHADOW_OUT)
            if dock: dock.hide()

    def OnToggleMenu(self, widget):
        if self.__menu:
            self.HideMenu()
        else:
            self.__arrow.set(gtk.ARROW_DOWN, gtk.SHADOW_IN)
            self.__menu = LaunchMenu(self.__menuitems, self, 0)
            if dock: dock.show()

class LaunchTip(RawWindow):
    def __init__(self, tip, widget, pos):
        RawWindow.__init__(self)
        self.set_name("FoalTooltip")
        f=gtk.GtkFrame()
        f.set_shadow_type(gtk.SHADOW_OUT)
        l=gtk.GtkLabel(tip)
        f.add(l)
        self.add(f)
        self.show_all()
        self.place_relative(widget, pos)

class LaunchMenu(RawWindow):

    def __init__(self, menu, widget, pos_flags):
        # - Determine max nos of menu per row/column
        if pos_flags & 4: layout = (len(menu),1)
        else: layout = (1, len(menu))
        sw, sh = gtk.screen_width()/(2*BUTTON_SIZE()), gtk.screen_height()/(2*BUTTON_SIZE())
        if layout[0] > sw: layout = ((layout[0]+1)/2, layout[1]*2)
        if layout[1] > sh: layout = (layout[0]*2, (layout[1]+1)/2)

        # - Create the menu window
        RawWindow.__init__(self, (BUTTON_SIZE()*layout[0], BUTTON_SIZE()*layout[1]))
        box = gtk.GtkTable(layout[0], layout[1], 0)
        box.set_row_spacings(0)
        box.set_col_spacings(0)

        # - Add an entry for each menu item
        r, c = 0, 0
        for entry in menu:
            button = gtk.GtkButton()
            button.add(self.load_icon(entry.icon, entry.name))
            button.set_usize(BUTTON_SIZE(), BUTTON_SIZE())
            button.set_data("command", entry)
            button.connect("clicked", self.OnLaunchCommand)
            if isinstance(entry, LaunchMenuItem):
                button.set_name("FoalMenuItem")
            else:
                button.set_name("FoalSubMenu")
            box.attach(button, c, c+1, r, r+1)
            c = c + 1
            if c == layout[0]:
                r = r + 1
                c = 0
        self.add(box)

        self.__submenu = None
        self.__submenu_name = None

        # - Place menu
        self.show_all()
        self.place_relative(widget, pos_flags)
        self.__pos_flags = pos_flags

        # - Start up dynamic tooltips
        self.__tip_timer = gtk.timeout_add(2000,
                                           self.NextTip)
        self.__tip_index = 0
        self.__tip = None
        self.__tips = box.children()[:]
        self.connect("destroy", self.StopTip)
        self.connect("destroy", self.KillSubMenu)

    def NextTip(self):
        if self.__tip:
            self.__tip.destroy()
            self.__tip = None
        if self.__tip_index >= len(self.__tips): return 0
        i = self.__tip_index
        self.__tip=LaunchTip(self.__tips[i].get_data("command").name, self.__tips[i], self.__pos_flags ^ 4)
        self.__tip_index = (i+1) % len(self.__tips)
        return 1

    def StopTip(self, widget=None):
        if self.__tip_timer:
            gtk.timeout_remove(self.__tip_timer)
            self.__tip_timer = None
        if self.__tip:
            self.__tip.destroy()
            self.__tip = None

    def KillSubMenu(self, widget=None):
        if self.__submenu:
            self.__submenu.destroy()
            self.__submenu = None
            self.__submenu_name = None

    def OnLaunchCommand(self, button):
        self.StopTip()
        submenu = self.__submenu_name
        self.KillSubMenu()
        item = button.get_data("command")
        if isinstance(item, LaunchMenuItem):
            if item.command: LaunchCommand(item.command)
        elif submenu != item.name:
            self.__submenu = LaunchMenu(item.items, button, self.__pos_flags ^ 4)
            self.__submenu_name = item.name

def LaunchCommand(command):
    if command[:6] == "#foal.":
        Foal(command[6:])
        return
    try:
        # - Is this the best way to launch an app?
        os.system(command + "&")
    except:
        sys.stderr.write("warning:failed to Launch "+command+"\n")

class LaunchMenuItem:
    def __init__(self, name, icon, program):
        self.name = name
        self.icon = icon
        self.command = program

class LaunchMenuSubMenu:
    def __init__(self, name, icon, items):
        self.name = name
        self.icon = icon
        self.items = items

# - Foal Dock bar.  This is ugly, due to Gtk+ queuing issues.

class LaunchDock(RawWindow):
    def __init__(self, launcher):
        RawWindow.__init__(self)
        #self.set_resize_mode(gtk.RESIZE_IMMEDIATE)
        box = gtk.GtkHButtonBox()
        box.set_spacing(0)
        box.set_child_ipadding(0, 0)
        box.set_child_size(BUTTON_SIZE(), BUTTON_SIZE())
        self.box = box
        self.add(box)
        self.show_all()
        self.hide()
    def DockWindow(self, xid):
        s = gtk.GtkSocket()
        f=gtk.GtkFrame()
        f.set_shadow_type(gtk.SHADOW_OUT)
        f.add(s)
        self.box.add(f)
        f.show_all()
        s.connect("destroy", self.UnDockWindow, f)
        s.steal(xid)
        w=len(self.box.children())*BUTTON_SIZE()
        self.set_usize(w, BUTTON_SIZE())
        self.place_relative(launcher, 4, (w, BUTTON_SIZE()))
        self.show_all()
        print "Dock:", f
    def UnDockWindow(self, widget, frame):
        print "UnDock:", frame
        frame.destroy()
        w=len(self.box.children())*BUTTON_SIZE()
        if w and launcher:
            self.set_usize(w, BUTTON_SIZE())
            self.place_relative(launcher, 4, (w, BUTTON_SIZE()))
        else:
            self.hide()

launcher = None
dock = None

# *** HACK
def ProcessEvents():
    while gtk.events_pending():
        gtk.mainiteration()

def Foal(func):
    global launcher, dock
    for f in string.split(func, ","):
        f2 = string.split(f)
        func = f2[0]
        if (func=="open") and not launcher:
            launcher = LaunchButton()
        elif (func=="close") and launcher:
            launcher.destroy()
            launcher = None
        elif func=="quit":
            Foal("dockclose,close")
            gtk.mainquit()
        elif (func=="dockopen") and launcher and not dock:
            ProcessEvents()
            dock = LaunchDock(launcher)
        elif (func=="dockclose") and dock:
            dock.destroy()
            dock = None
        elif (func=="dock") and dock:
            print "Running xwininfo..."
            tokens = string.split(os.popen("xwininfo").read())
            for i in range(len(tokens)):
                if tokens[i] == "id:":
                    print "Docking:", tokens[i+1]
                    dock.DockWindow(eval(tokens[i+1]))
                    break

def Main(argv):
    # - Create & run the Launcher
    gtk.rc_parse(os.path.join(os.environ["HOME"],
                              ".foal/foalgtkrc"))
    gtk.rc_parse("/etc/foal/foalgtkrc")
    Foal("open")
    gtk.mainloop()

if __name__ == "__main__": Main(sys.argv)


        
        
